@@ -1,31 +1,38 @@
 /****************************************************************************
 **
-** Copyright (C) 2015 The Qt Company Ltd.
-** Contact: http://www.qt.io/licensing/
+** Copyright (C) 2016 The Qt Company Ltd.
+** Copyright (C) 2016 Intel Corporation.
+** Contact: https://www.qt.io/licensing/
 **
 ** This file is part of the QtCore module of the Qt Toolkit.
 **
-** $QT_BEGIN_LICENSE:LGPL21$
+** $QT_BEGIN_LICENSE:LGPL$
 ** Commercial License Usage
 ** Licensees holding valid commercial Qt licenses may use this file in
 ** accordance with the commercial license agreement provided with the
 ** Software or, alternatively, in accordance with the terms contained in
 ** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see http://www.qt.io/terms-conditions. For further
-** information use the contact form at http://www.qt.io/contact-us.
+** and conditions see https://www.qt.io/terms-conditions. For further
+** information use the contact form at https://www.qt.io/contact-us.
 **
 ** GNU Lesser General Public License Usage
 ** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 2.1 or version 3 as published by the Free
-** Software Foundation and appearing in the file LICENSE.LGPLv21 and
-** LICENSE.LGPLv3 included in the packaging of this file. Please review the
-** following information to ensure the GNU Lesser General Public License
-** requirements will be met: https://www.gnu.org/licenses/lgpl.html and
-** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
+** General Public License version 3 as published by the Free Software
+** Foundation and appearing in the file LICENSE.LGPL3 included in the
+** packaging of this file. Please review the following information to
+** ensure the GNU Lesser General Public License version 3 requirements
+** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
 **
-** As a special exception, The Qt Company gives you certain additional
-** rights. These rights are described in The Qt Company LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
+** GNU General Public License Usage
+** Alternatively, this file may be used under the terms of the GNU
+** General Public License version 2.0 or (at your option) the GNU General
+** Public license version 3 or any later version approved by the KDE Free
+** Qt Foundation. The licenses are as published by the Free Software
+** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
+** included in the packaging of this file. Please review the following
+** information to ensure the GNU General Public License requirements will
+** be met: https://www.gnu.org/licenses/gpl-2.0.html and
+** https://www.gnu.org/licenses/gpl-3.0.html.
 **
 ** $QT_END_LICENSE$
 **
@@ -54,6 +61,7 @@
 #include <qendian.h>
 #include <qnumeric.h>
 
+#include "private/qendian_p.h"
 #include "private/qsimd_p.h"
 
 #include <limits.h>
@@ -61,6 +69,9 @@
 
 QT_BEGIN_NAMESPACE
 
+// in qstring.cpp
+void qt_to_latin1_unchecked(uchar *dst, const ushort *uc, qsizetype len);
+
 /*
   This defines a binary data structure for Json data. The data structure is optimised for fast reading
   and minimum allocations. The whole data structure can be mmap'ed and used directly.
@@ -115,6 +126,7 @@ QT_BEGIN_NAMESPACE
     Other measurements have shown a slightly bigger binary size than a compact text
     representation where all possible whitespace was stripped out.
 */
+#define Q_DECLARE_JSONPRIVATE_TYPEINFO(Class, Flags) } Q_DECLARE_TYPEINFO(QJsonPrivate::Class, Flags); namespace QJsonPrivate {
 namespace QJsonPrivate {
 
 class Array;
@@ -123,27 +135,7 @@ class Value;
 class Entry;
 
 template<typename T>
-class q_littleendian
-{
-public:
-    T val;
-
-    q_littleendian &operator =(T i) { val = qToLittleEndian(i); return *this; }
-    operator T() const { return qFromLittleEndian(val); }
-
-    bool operator ==(T i) { return qFromLittleEndian(val) == i; }
-    bool operator !=(T i) { return qFromLittleEndian(val) != i; }
-    bool operator ==(q_littleendian<T> i) { return val == i.val; }
-    bool operator !=(q_littleendian<T> i) { return val != i.val; }
-    bool operator <(T i) { return qFromLittleEndian(val) < i; }
-    bool operator >(T i) { return qFromLittleEndian(val) > i; }
-    bool operator <=(T i) { return qFromLittleEndian(val) <= i; }
-    bool operator >=(T i) { return qFromLittleEndian(val) >= i; }
-    q_littleendian &operator +=(T i) {
-        val = qToLittleEndian(qFromLittleEndian(val) + i);
-        return *this;
-    }
-};
+using q_littleendian = QLEInteger<T>;
 
 typedef q_littleendian<short> qle_short;
 typedef q_littleendian<unsigned short> qle_ushort;
@@ -151,89 +143,10 @@ typedef q_littleendian<int> qle_int;
 typedef q_littleendian<unsigned int> qle_uint;
 
 template<int pos, int width>
-class qle_bitfield
-{
-public:
-    uint val;
-
-    enum {
-        mask = ((1u << width) - 1) << pos
-    };
-
-    void operator =(uint t) {
-        uint i = qFromLittleEndian(val);
-        i &= ~mask;
-        i |= t << pos;
-        val = qToLittleEndian(i);
-    }
-    operator uint() const {
-        uint t = qFromLittleEndian(val);
-        t &= mask;
-        t >>= pos;
-        return t;
-    }
-    bool operator !() const {
-        return !operator uint();
-    }
-
-    bool operator ==(uint t) { return uint(*this) == t; }
-    bool operator !=(uint t) { return uint(*this) != t; }
-    bool operator <(uint t) { return uint(*this) < t; }
-    bool operator >(uint t) { return uint(*this) > t; }
-    bool operator <=(uint t) { return uint(*this) <= t; }
-    bool operator >=(uint t) { return uint(*this) >= t; }
-    qle_bitfield &operator +=(uint i) {
-        *this = (uint(*this) + i);
-        return *this;
-    }
-    qle_bitfield &operator -=(uint i) {
-        *this = (uint(*this) - i);
-        return *this;
-    }
-};
+using qle_bitfield = QLEIntegerBitfield<uint, pos, width>;
 
 template<int pos, int width>
-class qle_signedbitfield
-{
-public:
-    uint val;
-
-    enum {
-        mask = ((1u << width) - 1) << pos
-    };
-
-    void operator =(int t) {
-        uint i = qFromLittleEndian(val);
-        i &= ~mask;
-        i |= t << pos;
-        val = qToLittleEndian(i);
-    }
-    operator int() const {
-        uint i = qFromLittleEndian(val);
-        i <<= 32 - width - pos;
-        int t = (int) i;
-        t >>= pos;
-        return t;
-    }
-    bool operator !() const {
-        return !operator int();
-    }
-
-    bool operator ==(int t) { return int(*this) == t; }
-    bool operator !=(int t) { return int(*this) != t; }
-    bool operator <(int t) { return int(*this) < t; }
-    bool operator >(int t) { return int(*this) > t; }
-    bool operator <=(int t) { return int(*this) <= t; }
-    bool operator >=(int t) { return int(*this) >= t; }
-    qle_signedbitfield &operator +=(int i) {
-        *this = (int(*this) + i);
-        return *this;
-    }
-    qle_signedbitfield &operator -=(int i) {
-        *this = (int(*this) - i);
-        return *this;
-    }
-};
+using qle_signedbitfield = QLEIntegerBitfield<int, pos, width>;
 
 typedef qle_uint offset;
 
@@ -244,14 +157,7 @@ static inline bool useCompressed(const QString &s)
 {
     if (s.length() >= 0x8000)
         return false;
-    const ushort *uc = (const ushort *)s.constData();
-    const ushort *e = uc + s.length();
-    while (uc < e) {
-        if (*uc > 0xff)
-            return false;
-        ++uc;
-    }
-    return true;
+    return QtPrivate::isLatin1(s);
 }
 
 static inline int qStringSize(const QString &string, bool compress)
@@ -292,15 +198,22 @@ class Latin1String;
 class String
 {
 public:
-    String(const char *data) { d = (Data *)data; }
+    explicit String(const char *data) { d = (Data *)data; }
 
     struct Data {
-        qle_int length;
+        qle_uint length;
         qle_ushort utf16[1];
     };
 
     Data *d;
 
+    int byteSize() const { return sizeof(uint) + sizeof(ushort) * d->length; }
+    bool isValid(int maxSize) const {
+        // Check byteSize() <= maxSize, avoiding integer overflow
+        maxSize -= sizeof(uint);
+        return maxSize >= 0 && uint(d->length) <= maxSize / sizeof(ushort);
+    }
+
     inline String &operator=(const QString &str)
     {
         d->length = str.length();
@@ -366,67 +279,35 @@ public:
 class Latin1String
 {
 public:
-    Latin1String(const char *data) { d = (Data *)data; }
+    explicit Latin1String(const char *data) { d = (Data *)data; }
 
     struct Data {
-        qle_short length;
+        qle_ushort length;
         char latin1[1];
     };
     Data *d;
 
+    int byteSize() const { return sizeof(ushort) + sizeof(char)*(d->length); }
+    bool isValid(int maxSize) const {
+        return byteSize() <= maxSize;
+    }
+
     inline Latin1String &operator=(const QString &str)
     {
         int len = d->length = str.length();
         uchar *l = (uchar *)d->latin1;
         const ushort *uc = (const ushort *)str.unicode();
-        int i = 0;
-#ifdef __SSE2__
-        for ( ; i + 16 < len; i += 16) {
-            __m128i chunk1 = _mm_loadu_si128((__m128i*)&uc[i]); // load
-            __m128i chunk2 = _mm_loadu_si128((__m128i*)&uc[i + 8]); // load
-            // pack the two vector to 16 x 8bits elements
-            const __m128i result = _mm_packus_epi16(chunk1, chunk2);
-            _mm_storeu_si128((__m128i*)&l[i], result); // store
-        }
-#  ifdef Q_PROCESSOR_X86_64
-        // we can do one more round, of 8 characters
-        if (i + 8 < len) {
-            __m128i chunk = _mm_loadu_si128((__m128i*)&uc[i]); // load
-            // pack with itself, we'll discard the high part anyway
-            chunk = _mm_packus_epi16(chunk, chunk);
-            // unaligned 64-bit store
-            *(quint64*)&l[i] = _mm_cvtsi128_si64(chunk);
-            i += 8;
-        }
-#  endif
-#endif
-        for ( ; i < len; ++i)
-            l[i] = uc[i];
-        for ( ; (quintptr)(l+i) & 0x3; ++i)
-            l[i] = 0;
+        qt_to_latin1_unchecked(l, uc, len);
+
+        for ( ; (quintptr)(l+len) & 0x3; ++len)
+            l[len] = 0;
         return *this;
     }
 
-    inline bool operator ==(const QString &str) const {
-        return QLatin1String(d->latin1, d->length) == str;
-    }
-    inline bool operator !=(const QString &str) const {
-        return !operator ==(str);
-    }
-    inline bool operator >=(const QString &str) const {
-        return QLatin1String(d->latin1, d->length) >= str;
+    QLatin1String toQLatin1String() const Q_DECL_NOTHROW {
+        return QLatin1String(d->latin1, d->length);
     }
 
-    inline bool operator ==(const Latin1String &str) const {
-        return d->length == str.d->length && !strcmp(d->latin1, str.d->latin1);
-    }
-    inline bool operator >=(const Latin1String &str) const {
-        int l = qMin(d->length, str.d->length);
-        int val = strncmp(d->latin1, str.d->latin1, l);
-        if (!val)
-            val = d->length - str.d->length;
-        return val >= 0;
-    }
     inline bool operator<(const String &str) const
     {
         const qle_ushort *uc = (qle_ushort *) str.d->utf16;
@@ -457,6 +338,36 @@ public:
     }
 };
 
+#define DEF_OP(op) \
+    inline bool operator op(Latin1String lhs, Latin1String rhs) Q_DECL_NOTHROW \
+    { \
+        return lhs.toQLatin1String() op rhs.toQLatin1String(); \
+    } \
+    inline bool operator op(QLatin1String lhs, Latin1String rhs) Q_DECL_NOTHROW \
+    { \
+        return lhs op rhs.toQLatin1String(); \
+    } \
+    inline bool operator op(Latin1String lhs, QLatin1String rhs) Q_DECL_NOTHROW \
+    { \
+        return lhs.toQLatin1String() op rhs; \
+    } \
+    inline bool operator op(const QString &lhs, Latin1String rhs) Q_DECL_NOTHROW \
+    { \
+        return lhs op rhs.toQLatin1String(); \
+    } \
+    inline bool operator op(Latin1String lhs, const QString &rhs) Q_DECL_NOTHROW \
+    { \
+        return lhs.toQLatin1String() op rhs; \
+    } \
+    /*end*/
+DEF_OP(==)
+DEF_OP(!=)
+DEF_OP(< )
+DEF_OP(> )
+DEF_OP(<=)
+DEF_OP(>=)
+#undef DEF_OP
+
 inline bool String::operator ==(const Latin1String &str) const
 {
     if ((int)d->length != (int)str.d->length)
@@ -521,7 +432,7 @@ static inline void copyString(char *dest, const QString &str, bool compress)
 
 
 /*
- Base is the base class for both Object and Array. Both classe work more or less the same way.
+ Base is the base class for both Object and Array. Both classes work more or less the same way.
  The class starts with a header (defined by the struct below), then followed by data (the data for
  values in the Array case and Entry's (see below) for objects.
 
@@ -558,9 +469,10 @@ public:
     Entry *entryAt(int i) const {
         return reinterpret_cast<Entry *>(((char *)this) + table()[i]);
     }
-    int indexOf(const QString &key, bool *exists);
+    int indexOf(const QString &key, bool *exists) const;
+    int indexOf(QLatin1String key, bool *exists) const;
 
-    bool isValid() const;
+    bool isValid(int maxSize) const;
 };
 
 
@@ -570,7 +482,7 @@ public:
     inline Value at(int i) const;
     inline Value &operator [](int i);
 
-    bool isValid() const;
+    bool isValid(int maxSize) const;
 };
 
 
@@ -605,6 +517,7 @@ public:
     static uint valueToStore(const QJsonValue &v, uint offset);
     static void copyData(const QJsonValue &v, char *dest, bool compressed);
 };
+Q_DECLARE_JSONPRIVATE_TYPEINFO(Value, Q_PRIMITIVE_TYPE)
 
 inline Value Array::at(int i) const
 {
@@ -624,12 +537,12 @@ public:
     // key
     // value data follows key
 
-    int size() const {
+    uint size() const {
         int s = sizeof(Entry);
         if (value.latinKey)
-            s += sizeof(ushort) + qFromLittleEndian(*(ushort *) ((const char *)this + sizeof(Entry)));
+            s += shallowLatin1Key().byteSize();
         else
-            s += sizeof(uint) + sizeof(ushort)*qFromLittleEndian(*(int *) ((const char *)this + sizeof(Entry)));
+            s += shallowKey().byteSize();
         return alignedSize(s);
     }
 
@@ -655,10 +568,23 @@ public:
         return shallowKey().toString();
     }
 
+    bool isValid(int maxSize) const {
+        if (maxSize < (int)sizeof(Entry))
+            return false;
+        maxSize -= sizeof(Entry);
+        if (value.latinKey)
+            return shallowLatin1Key().isValid(maxSize);
+        return shallowKey().isValid(maxSize);
+    }
+
     bool operator ==(const QString &key) const;
     inline bool operator !=(const QString &key) const { return !operator ==(key); }
     inline bool operator >=(const QString &key) const;
 
+    bool operator==(QLatin1String key) const;
+    inline bool operator!=(QLatin1String key) const { return !operator ==(key); }
+    inline bool operator>=(QLatin1String key) const;
+
     bool operator ==(const Entry &other) const;
     bool operator >=(const Entry &other) const;
 };
@@ -671,9 +597,20 @@ inline bool Entry::operator >=(const QString &key) const
         return (shallowKey() >= key);
 }
 
+inline bool Entry::operator >=(QLatin1String key) const
+{
+    if (value.latinKey)
+        return shallowLatin1Key() >= key;
+    else
+        return shallowKey() >= key;
+}
+
 inline bool operator <(const QString &key, const Entry &e)
 { return e >= key; }
 
+inline bool operator<(QLatin1String key, const Entry &e)
+{ return e >= key; }
+
 
 class Header {
 public:
@@ -788,7 +725,11 @@ public:
         if (reserve) {
             if (reserve < 128)
                 reserve = 128;
-            size = qMax(size + reserve, size *2);
+            size = qMax(size + reserve, qMin(size *2, (int)Value::MaxSize));
+            if (size > Value::MaxSize) {
+                qWarning("QJson: Document too large to store in data structure");
+                return 0;
+            }
         }
         char *raw = (char *)malloc(size);
         Q_CHECK_PTR(raw);
@@ -805,7 +746,7 @@ public:
     bool valid() const;
 
 private:
-    Q_DISABLE_COPY(Data)
+    Q_DISABLE_COPY_MOVE(Data)
 };
 
 }
